import { Client, EmbedBuilder, Guild, HexColorString, Channel as DiscordChannel, TextChannel } from "discord.js";
import RssParser from "rss-parser";
import { parse as HtmlParser } from "node-html-parser";
import { Feed, Filter, MessageStyle, Channel, MatchingAlgorithms } from "../../generated/prisma";
import * as filters from "@bot/filter";
import prisma from "@server/prisma";

export const triggerTask = async (client: Client) => {
    for (const [_, guild] of client.guilds.cache) {
        await processGuild(guild, client);
    }
};

interface ExpandedFeed extends Feed {
    channels: Channel[],
    filters: Filter[],
    message_style: MessageStyle
}

const processGuild = async (guild: Guild, client: Client) => {
    const feeds = await prisma.feed.findMany({
        where: { guild_id: guild.id, active: true },
        include: { channels: true, filters: true, message_style: true }
    }) as ExpandedFeed[];

    for (const feed of feeds) {
        await processFeed(feed, client);
    }
};

const getParsedUrl = async (url: string) => {
    const parser = new RssParser();

    try { return parser.parseURL(url) }
    catch (error) { return undefined }
};

const processFeed = async (feed: ExpandedFeed, client: Client) => {
    const parsed = await getParsedUrl(feed.url);
    if (!parsed) return;

    for (const channelId of feed.channels.map(channel => channel.channel_id)) {
        const channel = client.channels.cache.get(channelId);
        if (channel) await processItems(parsed.items, feed, channel, client);
    }
};

const processItems = async (items: RssParser.Item[], feed: ExpandedFeed, channel: DiscordChannel, client: Client) => {
    for (const item of items) {
        await processItem(item, feed, channel, client);
    }
};

const processItem = async (item: RssParser.Item, feed: ExpandedFeed, channel: DiscordChannel, client: Client) => {
    const filterPromises = feed.filters.map(async filter => {
        return passesFilter(filter, item);
    });

    const filterResults = await Promise.all(filterPromises);
    if (filterResults.includes(false)) return;

    const embed = new EmbedBuilder();
    embed.setTitle(item.title ?? null);
    embed.setDescription(item.contentSnippet ?? null);
    embed.setImage(await getItemImageUrl(item.link ?? "") ?? null);
    embed.setColor(feed.message_style.colour as HexColorString);

    await (channel as TextChannel).send({ embeds: [embed] });
};

const getItemImageUrl = async (url: string) => {
    const response = await fetch(url);
    const html = HtmlParser.parse(await response.text());

    const imageElement = html.querySelector("meta[property='og:image']");
    if (!imageElement) return "";

    return imageElement.getAttribute("content");
};

const passesFilter = async (filter: Filter, item: RssParser.Item) => {
    if (!filter.matching_algorithm.trim()) return !filter.is_whitelist;

    let matchFound = false;

    if (filter.matching_algorithm === MatchingAlgorithms.ALL) {
        matchFound = filters.all(filter, `${item.title} ${item.content}`);
    } else {
        matchFound = (
            filters.mapAlgorithmToFunction(filter, item.title ?? "")
            || filters.mapAlgorithmToFunction(filter, item.content ?? "")
        );
    }

    return filter.is_whitelist ? matchFound : !matchFound;
};
